//======================================
//=== Copyright (c) 2014 Dušan Hupka ===
//=== BSD License                    ===
//=== xhupka00@stud.fit.vutbr.cz     ===
//======================================



#include "MyLOD.h"






MyLOD::MyLOD(Node* in_node)
{
	_history = new MyLODHistory();
	_cm =  new osgGA::TrackballManipulator();
	_texStore = new MyTextureStore();
	_billboarTexStore = new MyTextureStore();
	_isBillboard = false;
	_createdBillboard = false;

	if (in_node != NULL) {
		_mesh = dynamic_cast<Node*>(in_node->clone(osg::CopyOp::DEEP_COPY_ALL));

		_mesh->accept(*_texStore);
		_history->CreateHistoryLine("Original Mesh", _mesh);
		ComputePolyCount();
	}
}


MyLOD::~MyLOD()
{
	_mesh = NULL;
	delete _history;
}


void MyLOD::CopyCameraManipulator(osgGA::TrackballManipulator* A, osgGA::TrackballManipulator* B)
{
	A->setByMatrix(B->getMatrix());
	A->setCenter(B->getCenter());
	A->setDistance(B->getDistance());

}


QStringList MyLOD::GetMethodStrings() 
{ 
	QStringList rtv; 
	rtv.push_back("OSG_ORIGINAL_SIMPLIFIER"); 
	rtv.push_back("QUADRIC_ERROR_METRIC"); 
	rtv.push_back("SHORTEST_EDGE"); 
	rtv.push_back("SHORTEST_EDGE_NORMAL"); 
	rtv.push_back("EDGES_SHORTER_PIXEL"); 
	rtv.push_back("STAN MELAX"); 
	return rtv; 
}



bool MyLOD::IsUndoPossible() 
{
	return _history->IsUndoPossible();
}


bool MyLOD::IsRedoPossible() 
{
	return _history->IsRedoPossible();	
}


void MyLOD::Undo() 
{
	if (!_history->IsUndoPossible()) 
		return;
	_mesh = _history->Undo();
	ComputePolyCount();
}


void MyLOD::Redo() 
{
	if (!_history->IsRedoPossible()) 
		return;
	_mesh = _history->Redo();
	ComputePolyCount();
}

// simplifikacia geometrie, return: priemer hodnoty zvysnych hran po simplifikacii
float MyLOD::Simplify(int in_method, float in_ratio, float in_maxErr, bool in_edges) 
{

	Simplifier simp(in_ratio, in_maxErr);
	simp.setIgnoreEdges(in_edges);
	if (in_method == QUADRIC_ERROR_MATRIC) {
		simp.setErrorMetric(Simplifier::QEM);
	}	
	else if (in_method == STAN_MELAX) {
		simp.setErrorMetric(Simplifier::STAN_MELAX);
	}
	else if (in_method == SHORTEST_EDGE) {
		simp.setErrorMetric(Simplifier::SHORTEST_EDGE);
	}
	else if (in_method == SHORTEST_EDGE_NORMAL) {
		simp.setErrorMetric(Simplifier::SHORTEST_EDGE_NORMAL);
	}
	else if (in_method == EDGES_SHORTER_PIXEL) {
		simp.setErrorMetric(Simplifier::SHORTEST_EDGE);
		simp.setSampleRatio(0.0001);
	}
	_mesh->accept(simp);

	stringstream hstr;
	hstr << "Apply " << MyLOD::GetMethodStrings()[in_method].toStdString() << " with ratio " << in_ratio << " and MaxError " << in_maxErr;
	_history->CreateHistoryLine(hstr.str(), _mesh);
	ComputePolyCount();
	return simp.getMeanEdgeError();
}


osgGA::TrackballManipulator* MyLOD::GetCameraManipulator() 
{	
	return _cm;
}


float MyLOD::GetLODDistance() 
{
	return (float)_cm->getDistance();	 
}


float MyLOD::GetLODPixels() 
{
	return _pixelSize;
}


void MyLOD::SetLODPixels(float in_size)
{
	_pixelSize = in_size;
}


void MyLOD::ComputePolyCount()
{
	MyVertCount pc;
	_mesh->accept(pc);
	_polyCount = pc.GetPolyCount();
}


long MyLOD::GetPolyCount()
{
	if (isBillboard())
		return 2;
	return _polyCount;
}


Node* MyLOD::GetNode()
{
	if (isBillboard())
		return _billboardGrp;
	else
		return _mesh;
}

MyTextureStore* MyLOD::GetTextureStore()
{
	if (isBillboard())
		return _billboarTexStore;
	else 
		return _texStore;
}


bool MyLOD::isBillboard()
{
	return _isBillboard;
}


void MyLOD::SetActiveBillboard(bool in_val)
{
	if (in_val && !_createdBillboard)
		CreateBillboard();
	_isBillboard = in_val;
}


// vytvori prvy plane
Geometry* MyLOD::CreatePlane()
{
	osg::Geometry* geom = new osg::Geometry;
	
	osg::Vec3Array* verts = new osg::Vec3Array(4);
	(*verts)[0] = osg::Vec3(-0.5f, 0, -0.5f);
	(*verts)[1] = (*verts)[0] + osg::Vec3(1,0,0); //width
	(*verts)[2] = (*verts)[0] + osg::Vec3(1,0,1);
	(*verts)[3] = (*verts)[0] + osg::Vec3(0,0,1); //height
	
	geom->setVertexArray(verts);
	
	osg::Vec3Array* norms = new osg::Vec3Array(4);
	(*norms)[0] = osg::Vec3(1,0,0) ^ osg::Vec3(0,1,0);
	(*norms)[0].normalize();
	(*norms)[1] = (*norms)[0];
	(*norms)[2] = (*norms)[0];
	(*norms)[3] = (*norms)[0];
	
	geom->setNormalArray(norms, osg::Array::BIND_PER_VERTEX);
	
	osg::Vec2Array* tcoords = new osg::Vec2Array(4);
	(*tcoords)[0].set(0.0f,0.0f);
	(*tcoords)[1].set(1.0f,0.0f);
	(*tcoords)[2].set(1.0f,1.0f);
	(*tcoords)[3].set(0.0f,1.0f);
	geom->setTexCoordArray(0, tcoords, osg::Array::BIND_PER_VERTEX);
	
	geom->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::QUADS,0,4));

	return geom;
}


// podla axis prepocita novy plane pre billboard
void MyLOD::RebuildBillboardPlaneByAxis(EBillboardType in_mode)
{
	osg::Vec3Array* verts = (osg::Vec3Array*)_billboardGeom->getVertexArray();
	osg::Vec3Array* norms = (osg::Vec3Array*)_billboardGeom->getNormalArray();

	//X || Z
	if (in_mode == X_AXIS || in_mode == Z_AXIS) {
		(*verts)[0] = osg::Vec3(-0.5f, 0, -0.5f);
		(*verts)[1] = (*verts)[0] + osg::Vec3(1,0,0);
		(*verts)[2] = (*verts)[0] + osg::Vec3(1,0,1);
		(*verts)[3] = (*verts)[0] + osg::Vec3(0,0,1);
		(*norms)[0] = osg::Vec3(1,0,0) ^ osg::Vec3(0,0,1);
	}
	//Y
	else {
		(*verts)[0] = osg::Vec3(0, -0.5f, -0.5f);
		(*verts)[1] = (*verts)[0] + osg::Vec3(0,1,0);
		(*verts)[2] = (*verts)[0] + osg::Vec3(0,1,1);
		(*verts)[3] = (*verts)[0] + osg::Vec3(0,0,1);
		(*norms)[0] = osg::Vec3(0,1,0) ^ osg::Vec3(0,0,1);
	}
	
	(*norms)[0].normalize();
	(*norms)[1] = (*norms)[0];
	(*norms)[2] = (*norms)[0];
	(*norms)[3] = (*norms)[0];

	_billboardGeom->setVertexArray(verts);
	_billboardGeom->setNormalArray(norms, osg::Array::BIND_PER_VERTEX);
}


void MyLOD::CreateBillboard()
{
	_billboardGrp = new osg::Group();
	_billboardPos = new PositionAttitudeTransform();
	_billboard = new osg::Billboard();	

	_billboardGeom = CreatePlane();
	_billboard->addDrawable(_billboardGeom);
	_billboardPos->addChild(_billboard);
	_billboardGrp->addChild(_billboardPos);

	SetBillboardMode(Z_AXIS);

	_billboardPos->setPosition(_mesh.get()->getBound().center());

	_billboardSS = _billboard->getOrCreateStateSet();
	_billboardSS->setRenderingHint( osg::StateSet::TRANSPARENT_BIN );	
	_billboardSS->setMode(GL_BLEND,osg::StateAttribute::ON);


	_createdBillboard = true;
	_isBillboard = true;
}


void MyLOD::SetBillboardScale(float in_width, float in_height)
{
	if (!_billboardPos)
		return;
	Vec3 scale = _billboardPos->getScale();
	_billboardPos->setScale(Vec3(scale.x() * in_width, scale.y() * in_width, scale.z() * in_height));
}


void MyLOD::ResetBillboardScale()
{
	if (!_billboardPos)
		return;
	_billboardPos->setScale(Vec3(1, 1, 1));
}


void MyLOD::SetBillboardImage(std::string in_file)
{
	osg::Image* img = osgDB::readImageFile(in_file);
	if (!img || !img->valid())
		return;

	osg::Texture2D* texture = new osg::Texture2D();
	texture->setImage(img);
	_billboardSS->setTextureAttributeAndModes(0, texture, osg::StateAttribute::ON);
	_billboardSS->setMode(GL_LIGHTING,osg::StateAttribute::OFF);
	_billboarTexStore->Clear();
	GetNode()->accept(*_billboarTexStore);
}


void MyLOD::RotateBillboardTexCoords()
{
	osg::Vec2Array* tcoords = (osg::Vec2Array*)_billboardGeom->getTexCoordArray(0);
	Vec2 tmp = (*tcoords)[0];
	(*tcoords)[0] = (*tcoords)[1];
	(*tcoords)[1] = (*tcoords)[2];
	(*tcoords)[2] = (*tcoords)[3];
	(*tcoords)[3] = tmp;
	_billboardGeom->setTexCoordArray(0, tcoords);
}


void MyLOD::FlipBillboardTexCoords()
{
	osg::Vec2Array* tcoords = (osg::Vec2Array*)_billboardGeom->getTexCoordArray(0);
	Vec2 tmp1 = (*tcoords)[0];
	Vec2 tmp2 = (*tcoords)[3];
	(*tcoords)[0] = (*tcoords)[1];
	(*tcoords)[3] = (*tcoords)[2];
	(*tcoords)[1] = tmp1;
	(*tcoords)[2] = tmp2;
	_billboardGeom->setTexCoordArray(0, tcoords);
}

//nastavi billboard mode a prisposoby tomu normaly, axis a prepocita novy plane
void MyLOD::SetBillboardMode(EBillboardType in_mode)
{
	_billboardMode = in_mode;
	if (in_mode == POINT_ROT_EYE) {
		_billboard->setMode(Billboard::POINT_ROT_EYE);
	}
	else {
		_billboard->setMode(Billboard::AXIAL_ROT);
		RebuildBillboardPlaneByAxis(in_mode);
		if (in_mode == X_AXIS) {			
			_billboard->setAxis(Vec3(0,-1, 0));
			_billboard->setNormal(osg::Vec3(0.0f, -1.0f, 0.0f));
		}
		else if (in_mode == Y_AXIS) {
			_billboard->setAxis(Vec3(1, 0, 0));
			_billboard->setNormal(osg::Vec3(1.0f, 0.0f, 0.0f));
		}
		else if (in_mode == Z_AXIS) {
			_billboard->setAxis(Vec3(0,-1, 0));
			_billboard->setAxis(Vec3(0, 0, 1));
		}
	}
	
}
